#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025-2026 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Convert values HDF5/netCDF4 attribute."""

from __future__ import annotations

__all__ = ["adjust_attr"]

import numpy as np


# - main function ------------------------------------
def adjust_attr(dtype: str, attr_key: str, attr_val: np.generic) -> np.generic:
    """Return attribute converted to the same data type as its variable.

    Parameters
    ----------
    dtype :  str
      numpy data-type of variable
    attr_key :  str
      name of the attribute
    attr_val :  np.generic
      original value of the attribute

    Returns
    -------
    attr_val converted to dtype
    """
    if attr_key == "flag_values":
        return np.array(attr_val, dtype=dtype)

    if attr_key == "flag_masks":
        return np.array(attr_val, dtype=dtype)

    if attr_key == "scale_factor" and isinstance(attr_val, str):
        return eval(attr_val)

    if attr_key in ("valid_min", "valid_max", "valid_range"):
        match dtype:
            case "i1":
                res = np.int8(attr_val)
            case "i2":
                res = np.int16(attr_val)
            case "i4":
                res = np.int32(attr_val)
            case "i8":
                res = np.int64(attr_val)
            case "u1":
                res = np.uint8(attr_val)
            case "u2":
                res = np.uint16(attr_val)
            case "u4":
                res = np.uint32(attr_val)
            case "u8":
                res = np.uint64(attr_val)
            case "f2":
                res = np.float16(attr_val)
            case "f4":
                res = np.float32(attr_val)
            case "f8":
                res = np.float64(attr_val)
            case _:
                res = attr_val

        return res

    return attr_val



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test module for h5yaml class `NcYaml`."""

from __future__ import annotations

from importlib.resources import files
from pathlib import Path, PurePosixPath

import numpy as np
import pytest

from h5yaml.yaml_nc import NcYaml


class TestNcYaml:
    """Class to test NcYaml from h5yaml.yaml_nc."""

    NC_DEF = NcYaml(files("h5yaml.Data") / "nc_testing.yaml").nc_def
    FID = NcYaml(files("h5yaml.Data") / "nc_testing.yaml").diskless()

    def test_exceptions(self: TestNcYaml) -> None:
        """Unit-test for the class exdeptions."""
        # tests which should raise an exception because the file can not be created
        l1a_name = "/this/folder/does/not/exists/test.nc"
        with pytest.raises(RuntimeError, match=r"failed to create .*") as excinfo:
            NcYaml(files("h5yaml.Data") / "nc_testing.yaml").create(l1a_name)
        assert f"failed to create {l1a_name}" in str(excinfo)

        l1a_name = Path("/this/folder/does/not/exists/test.nc")
        with pytest.raises(RuntimeError, match=r"failed to create .*") as excinfo:
            NcYaml(files("h5yaml.Data") / "nc_testing.yaml").create(l1a_name)
        assert f"failed to create {l1a_name}" in str(excinfo)

        l1a_name = "/test.nc"
        with pytest.raises(RuntimeError, match=r"failed to create .*") as excinfo:
            NcYaml(files("h5yaml.Data") / "nc_testing.yaml").create(l1a_name)
        assert f"failed to create {l1a_name}" in str(excinfo.value)

        # tests which raise an exception because the YAML file is corrupted
        file_path = Path("nc_testing.yaml")
        with pytest.raises(FileNotFoundError, match=r".* not found") as excinfo:
            _ = NcYaml(file_path).nc_def
        assert f"{file_path} not found" in str(excinfo.value)

    def test_groups(self: TestNcYaml) -> None:
        """Unit-test to check the groups."""
        if "groups" not in self.NC_DEF:
            return

        for key in self.NC_DEF["groups"]:
            pkey = PurePosixPath(key)
            if len(pkey.parts) > 1:
                assert pkey.name in self.FID[pkey.parent].groups
            else:
                assert key in self.FID.groups

    def test_dimensions(self: TestNcYaml) -> None:
        """Unit-test to check the dimensions."""
        if "dimensions" not in self.NC_DEF:
            return

        for key in self.NC_DEF["dimensions"]:
            pkey = PurePosixPath(key)
            dim_grp = self.FID[pkey.parent] if pkey.is_absolute() else self.FID
            dim_name = pkey.name if pkey.is_absolute() else key
            assert dim_name in dim_grp.dimensions

            if dim_name in dim_grp.variables:
                nc_dim = dim_grp.variables[dim_name]
                if "_dtype" in self.NC_DEF["dimensions"][key]:
                    dtype = self.NC_DEF["dimensions"][key]["_dtype"]
                    assert nc_dim.dtype == str if dtype == "str" else dtype
                if "_size" in self.NC_DEF["dimensions"][key]:
                    assert nc_dim.size == self.NC_DEF["dimensions"][key]["_size"]
                if "_values" in self.NC_DEF["dimensions"][key]:
                    assert np.array_equal(
                        nc_dim[:], self.NC_DEF["dimensions"][key]["_values"]
                    )
                if "_range" in self.NC_DEF["dimensions"][key]:
                    assert np.array_equal(
                        nc_dim[:], np.arange(*self.NC_DEF["dimensions"][key]["_range"])
                    )
                for attr in self.NC_DEF["dimensions"][key]:
                    if attr[0] == "_":
                        continue

                    assert getattr(nc_dim, attr) == self.NC_DEF["dimensions"][key][attr]

    def test_compounds(self: TestNcYaml) -> None:
        """Unit-test to check the compounds."""
        if "compounds" not in self.NC_DEF:
            return

        for key in self.NC_DEF["compounds"]:
            assert key in self.FID.cmptypes

    def test_variables(self: TestNcYaml) -> None:
        """Unit-test to check the variables."""
        if "variables" not in self.NC_DEF:
            return

        for key in self.NC_DEF["variables"]:
            pkey = PurePosixPath(key)
            var_grp = self.FID[pkey.parent] if pkey.is_absolute() else self.FID
            var_name = pkey.name if pkey.is_absolute() else key
            assert var_name in var_grp.variables
            nc_var = var_grp.variables[var_name]

            if "_size" in self.NC_DEF["variables"][key]:
                assert nc_var.size == self.NC_DEF["variables"][key]["_size"]

            for attr in self.NC_DEF["variables"][key]:
                if attr[0] == "_":
                    continue

                if isinstance(
                    self.NC_DEF["variables"][key][attr], list | tuple | np.ndarray
                ):
                    assert np.array_equal(
                        getattr(nc_var, attr), self.NC_DEF["variables"][key][attr]
                    )
                else:
                    assert getattr(nc_var, attr) == self.NC_DEF["variables"][key][attr]

    def test_close(self: TestNcYaml) -> None:
        """Close the in-memory netCDF4 file."""
        self.FID.close()



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test module for h5yaml class `H5Yaml`."""

from __future__ import annotations

from importlib.resources import files
from pathlib import Path

import numpy as np
import pytest

from h5yaml.yaml_h5 import H5Yaml


class TestH5Yaml:
    """Class to test H5Yaml from h5yaml.yaml_h5."""

    H5_DEF = H5Yaml(files("h5yaml.Data") / "h5_testing.yaml").h5_def
    FID = H5Yaml(files("h5yaml.Data") / "h5_testing.yaml").diskless()

    def test_exceptions(self: TestH5Yaml) -> None:
        """Unit-test for class exeptions."""
        # tests which should raise an exception because the file can not be created
        l1a_name = "/this/folder/does/not/exists/test.h5"
        with pytest.raises(FileNotFoundError, match=r"[Errno 2] .*") as excinfo:
            H5Yaml(files("h5yaml.Data") / "h5_testing.yaml").create(l1a_name)
        assert "'No such file or directory" in str(excinfo)

        l1a_name = Path("/this/folder/does/not/exists/test.h5")
        with pytest.raises(FileNotFoundError, match=r"[Errno 2] .*") as excinfo:
            H5Yaml(files("h5yaml.Data") / "h5_testing.yaml").create(l1a_name)
        assert "'No such file or directory" in str(excinfo)

        l1a_name = "/test.h5"
        with pytest.raises(RuntimeError, match=r"failed create .*") as excinfo:
            H5Yaml(files("h5yaml.Data") / "h5_testing.yaml").create(l1a_name)
        assert f"failed create {l1a_name}" in str(excinfo.value)

        # tests which raise an exception because the YAML file is corrupted
        file_path = Path("h5_testing.yaml")
        with pytest.raises(FileNotFoundError, match=r".* not found") as excinfo:
            _ = H5Yaml(file_path).h5_def
        assert f"{file_path} not found" in str(excinfo.value)

        with pytest.raises(ValueError, match=r".* unlimited dimension") as excinfo:
            H5Yaml(files("h5yaml.Data") / "h5_unsupported.yaml").create("klad.h5")
        assert "has more than one unlimited dimension" in str(excinfo.value)

    def test_groups(self: TestH5Yaml) -> None:
        """Unit-test to check the groups."""
        if "groups" not in self.H5_DEF:
            return

        for key in self.H5_DEF["groups"]:
            assert key in self.FID

    def test_dimensions(self: TestH5Yaml) -> None:
        """Unit-test to check the dimensions."""
        if "dimensions" not in self.H5_DEF:
            return

        for key in self.H5_DEF["dimensions"]:
            assert key in self.FID
            if "_dtype" in self.H5_DEF["dimensions"][key]:
                dtype = self.H5_DEF["dimensions"][key]["_dtype"]
                assert self.FID[key].dtype == np.dtype("O") if dtype == "str" else dtype
            if "_size" in self.H5_DEF["dimensions"][key]:
                assert self.FID[key].size == self.H5_DEF["dimensions"][key]["_size"]
            if "_values" in self.H5_DEF["dimensions"][key]:
                assert np.array_equal(
                    [x.decode() for x in self.FID[key][:]],
                    self.H5_DEF["dimensions"][key]["_values"],
                )
            if "_range" in self.H5_DEF["dimensions"][key]:
                assert np.array_equal(
                    self.FID[key][:],
                    np.arange(*self.H5_DEF["dimensions"][key]["_range"]),
                )
            for attr in self.H5_DEF["dimensions"][key]:
                if attr[0] == "_":
                    continue
                assert self.FID[key].attrs[attr] == self.H5_DEF["dimensions"][key][attr]

    def test_compounds(self: TestH5Yaml) -> None:
        """Unit-test to check the compounds."""
        if "compounds" not in self.H5_DEF:
            return

        for key in self.H5_DEF["compounds"]:
            assert key in self.FID

    def test_variables(self: TestH5Yaml) -> None:
        """Unit-test to check the variables."""
        if "variables" not in self.H5_DEF:
            return

        dset = self.H5_DEF["variables"]
        for key in dset:
            assert key in self.FID
            # if "_dtype" in self.H5_DEF["variables"][key]:
            #     assert self.FID[key].dtype == self.H5_DEF["variables"][key]["_dtype"]
            if "_size" in dset[key]:
                assert self.FID[key].size == dset[key]["_size"]
            for attr in self.H5_DEF["variables"][key]:
                if attr[0] == "_":
                    continue

                if isinstance(dset[key][attr], list | tuple | np.ndarray):
                    assert np.array_equal(self.FID[key].attrs[attr], dset[key][attr])
                else:
                    assert self.FID[key].attrs[attr] == dset[key][attr]

    def test_close(self: TestH5Yaml) -> None:
        """Close the in-memory HDF5 file."""
        self.FID.close()



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025-2026 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Create HDF5/netCDF4 formatted file from a YAML configuration file using h5py."""

from __future__ import annotations

__all__ = ["H5Yaml"]

import logging
from pathlib import Path

import h5py
import numpy as np

from .conf_from_yaml import conf_from_yaml
from .lib.adjust_attr import adjust_attr

H5_LIBVER = ("v110", "latest")


# - class definition -----------------------------------
class H5Yaml:
    """Class to create a HDF5/netCDF4 formated file from a YAML configuration file.

    Parameters
    ----------
    h5_yaml_fl :  Path | str | list[Path | str]
       YAML files with the HDF5 format definition

    """

    def __init__(self: H5Yaml, h5_yaml_fl: Path | str | list[Path | str]) -> None:
        """Construct a H5Yaml instance."""
        self.logger = logging.getLogger("h5yaml.H5Yaml")
        self._h5_def = {
            "groups": set(),
            "attrs_global": {},
            "attrs_groups": {},
            "compounds": {},
            "dimensions": {},
            "variables": {},
        }

        for yaml_fl in h5_yaml_fl if isinstance(h5_yaml_fl, list) else [h5_yaml_fl]:
            try:
                config = conf_from_yaml(yaml_fl)
            except RuntimeError as exc:
                raise RuntimeError from exc

            for key in self._h5_def:
                if key in config:
                    self._h5_def[key] |= (
                        set(config[key]) if key == "groups" else config[key]
                    )

    def __attrs(self: H5Yaml, fid: h5py.File) -> None:
        """Create global and group attributes.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'r+')

        """
        for key, value in self._h5_def["attrs_global"].items():
            if key not in fid.attrs and value != "TBW":
                fid.attrs[key] = value

        for key, value in self._h5_def["attrs_groups"].items():
            if key not in fid.attrs and value != "TBW":
                fid[str(Path(key).parent)].attrs[Path(key).name] = value

    def __groups(self: H5Yaml, fid: h5py.File) -> None:
        """Create groups in HDF5 product.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'r+')

        """
        for key in self._h5_def["groups"]:
            if key not in fid:
                _ = fid.create_group(key, track_order=True)

    def __dimensions(self: H5Yaml, fid: h5py.File) -> None:
        """Add dimensions to HDF5 product.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'r+')

        """
        for key, val in self._h5_def["dimensions"].items():
            fillvalue = None
            if "_FillValue" in val:
                fillvalue = (
                    np.nan if val["_FillValue"] == "NaN" else int(val["_FillValue"])
                )

            if val["_size"] == 0:
                ds_chunk = val.get("_chunks")
                if ds_chunk is not None and not isinstance(ds_chunk, bool):
                    ds_chunk = tuple(ds_chunk)
                dset = fid.create_dataset(
                    key,
                    shape=(0,),
                    dtype="T" if val["_dtype"] == "str" else val["_dtype"],
                    chunks=ds_chunk,
                    maxshape=(None,),
                    fillvalue=fillvalue,
                )
            else:
                dset = fid.create_dataset(
                    key,
                    shape=(val["_size"],),
                    dtype="T" if val["_dtype"] == "str" else val["_dtype"],
                )
                if "_values" in val:
                    dset[:] = val["_values"]
                elif "_range" in val:
                    dset[:] = np.arange(*val["_range"], dtype=val["_dtype"])

            dset.make_scale(
                Path(key).name
                if "long_name" in val
                else "This is a netCDF dimension but not a netCDF variable."
            )
            for attr, attr_val in val.items():
                if not attr.startswith("_"):
                    dset.attrs[attr] = adjust_attr(val["_dtype"], attr, attr_val)

    def __compounds(self: H5Yaml, fid: h5py.File) -> None:
        """Add compound datatypes to HDF5 product.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'r+')

        """
        for key, val in self._h5_def["compounds"].items():
            fid[key] = np.dtype([(k, v[0]) for k, v in val.items()])

    def __variables(self: H5Yaml, fid: h5py.File) -> None:
        """Add datasets to HDF5 product.

        Parameters
        ----------
        fid :  h5py.File
           HDF5 file pointer (mode 'r+')

        """

        def add_compound_attr() -> None:
            compound = self._h5_def["compounds"][val["_dtype"]]
            res = [v[2] for k, v in compound.items() if len(v) == 3]
            if res:
                dset.attrs["units"] = [v[1] for k, v in compound.items()]
                dset.attrs["names"] = res
            else:
                dset.attrs["names"] = [v[1] for k, v in compound.items()]

        for key, val in self._h5_def["variables"].items():
            if val["_dtype"] in fid:  # True when compound-dtype
                is_compound = True
                ds_dtype = fid[val["_dtype"]].dtype
            else:
                is_compound = False
                ds_dtype = "T" if val["_dtype"] == "str" else val["_dtype"]

            fillvalue = None
            if "_FillValue" in val:
                fillvalue = (
                    np.nan if val["_FillValue"] == "NaN" else int(val["_FillValue"])
                )

            # check for scalar dataset
            if val["_dims"][0] == "scalar":
                dset = fid.create_dataset(key, (), dtype=ds_dtype, fillvalue=fillvalue)
                for attr, attr_val in val.items():
                    if attr.startswith("_"):
                        continue
                    dset.attrs[attr] = adjust_attr(val["_dtype"], attr, attr_val)

                if is_compound:
                    add_compound_attr()
                continue

            n_udim = 0
            ds_shape = ()
            ds_maxshape = ()
            for coord in val["_dims"]:
                dim_sz = fid[coord].size
                n_udim += int(dim_sz == 0)
                ds_shape += (dim_sz,)
                ds_maxshape += (dim_sz if dim_sz > 0 else None,)

            # currently, we can not handle more than one unlimited dimension
            if n_udim > 1:
                raise ValueError(f"{key} has more than one unlimited dimension")

            if None in ds_maxshape and val.get("_chunks") == "contiguous":
                raise KeyError(
                    "you can not create a contiguous dataset with unlimited dimensions."
                )

            # create the variable
            if val.get("_chunks") == "contiguous":
                dset = fid.create_dataset(
                    key,
                    ds_shape,
                    dtype=ds_dtype,
                    chunks=None,
                    maxshape=None,
                    fillvalue=fillvalue,
                )
            else:
                ds_chunk = val.get("_chunks")
                if ds_chunk is not None and not isinstance(ds_chunk, bool):
                    ds_chunk = tuple(ds_chunk)
                compression = None
                shuffle = False
                # currently only gzip compression is supported
                if "_compression" in val:
                    compression = val["_compression"]
                    shuffle = True

                if val.get("_vlen"):
                    ds_name = (
                        val["_dtype"].split("_")[0]
                        if "_" in val["_dtype"]
                        else val["_dtype"]
                    ) + "_vlen"
                    if ds_name not in fid:
                        fid[ds_name] = h5py.vlen_dtype(ds_dtype)
                    ds_dtype = fid[ds_name]
                    fillvalue = None

                dset = fid.create_dataset(
                    key,
                    ds_shape,
                    dtype=ds_dtype,
                    chunks=ds_chunk,
                    maxshape=ds_maxshape,
                    fillvalue=fillvalue,
                    compression=compression,
                    shuffle=shuffle,
                )

            for ii, coord in enumerate(val["_dims"]):
                dset.dims[ii].attach_scale(fid[coord])

            for attr, attr_val in val.items():
                if not attr.startswith("_"):
                    dset.attrs[attr] = adjust_attr(val["_dtype"], attr, attr_val)

            if is_compound:
                add_compound_attr()

    @property
    def h5_def(self: H5Yaml) -> dict:
        """Return definition of the HDF5/netCDF4 product."""
        return self._h5_def

    def diskless(self: H5Yaml) -> h5py.File:
        """Create a HDF5/netCDF4 file in memory."""
        fid = h5py.File.in_memory(track_order=True, libver=H5_LIBVER)
        self.__groups(fid)
        self.__dimensions(fid)
        self.__compounds(fid)
        self.__variables(fid)
        self.__attrs(fid)
        return fid

    def create(self: H5Yaml, l1a_name: Path | str) -> None:
        """Create a HDF5/netCDF4 file (overwrite if exist).

        Parameters
        ----------
        l1a_name :  Path | str
           Full name of the HDF5/netCDF4 file to be generated

        """
        try:
            with h5py.File(l1a_name, "w", track_order=True, libver=H5_LIBVER) as fid:
                self.__groups(fid)
                self.__dimensions(fid)
                self.__compounds(fid)
                self.__variables(fid)
                self.__attrs(fid)
        except PermissionError as exc:
            raise RuntimeError(f"failed create {l1a_name}") from exc



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Read settings from a YAML file."""

from __future__ import annotations

__all__ = ["conf_from_yaml"]

from pathlib import Path

import yaml


# - main function -----------------------------------
def conf_from_yaml(file_path: Path | str) -> dict:
    """Read settings from a YAML file: `file_path`.

    Parameters
    ----------
    file_path :  Path | str
       full path of YAML file

    Returns
    -------
    dict
       content of the configuration file

    """
    if isinstance(file_path, str):
        file_path = Path(file_path)

    if not file_path.is_file():
        raise FileNotFoundError(f"{file_path} not found")

    with file_path.open("r", encoding="ascii") as fid:
        try:
            settings = yaml.safe_load(fid)
        except yaml.parser.ParserError as exc:
            raise RuntimeError(f"Failed to parse {file_path}") from exc

    return settings



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025-2026 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Create netCDF4 formatted file from a YAML configuration file using netCDF4."""

from __future__ import annotations

__all__ = ["NcYaml"]

import logging
from pathlib import Path, PurePosixPath

import numpy as np

# pylint: disable=no-name-in-module
from netCDF4 import Dataset

from .conf_from_yaml import conf_from_yaml
from .lib.adjust_attr import adjust_attr


# - class definition -----------------------------------
class NcYaml:
    """Class to create a netCDF4 formated file from a YAML configuration file.

    Parameters
    ----------
    nc_yaml_fl :  Path | str | list[Path | str]
       YAML files with the netCDF4 format definition

    """

    def __init__(self: NcYaml, nc_yaml_fl: Path | str | list[Path | str]) -> None:
        """Construct a NcYaml instance."""
        self.logger = logging.getLogger("h5yaml.NcYaml")
        self._nc_def = {
            "groups": set(),
            "attrs_global": {},
            "attrs_groups": {},
            "compounds": {},
            "dimensions": {},
            "variables": {},
        }

        for yaml_fl in nc_yaml_fl if isinstance(nc_yaml_fl, list) else [nc_yaml_fl]:
            try:
                config = conf_from_yaml(yaml_fl)
            except RuntimeError as exc:
                raise RuntimeError from exc

            for key in self._nc_def:
                if key in config:
                    self._nc_def[key] |= (
                        set(config[key]) if key == "groups" else config[key]
                    )

    def __attrs(self: NcYaml, fid: Dataset) -> None:
        """Create global and group attributes.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        for key, val in self.nc_def["attrs_global"].items():
            if val == "TBW":
                continue

            if isinstance(val, str):
                fid.setncattr_string(key, val)
            else:
                fid.setncattr(key, val)

        for key, val in self.nc_def["attrs_groups"].items():
            if val == "TBW":
                continue

            if isinstance(val, str):
                fid[str(Path(key).parent)].setncattr_string(Path(key).name, val)
            else:
                fid[str(Path(key).parent)].setncattr(Path(key).name, val)

    def __groups(self: NcYaml, fid: Dataset) -> None:
        """Create groups in a netCDF4 product.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        for key in self.nc_def["groups"]:
            _ = fid.createGroup(key)

    def __dimensions(self: NcYaml, fid: Dataset) -> None:
        """Add dimensions to a netCDF4 product.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        for key, value in self.nc_def["dimensions"].items():
            pkey = PurePosixPath(key)
            if pkey.is_absolute():
                _ = fid[pkey.parent].createDimension(pkey.name, value["_size"])
            else:
                _ = fid.createDimension(key, value["_size"])

            if len(value) <= 2:
                continue

            fillvalue = None
            if "_FillValue" in value:
                fillvalue = (
                    np.nan if value["_FillValue"] == "NaN" else int(value["_FillValue"])
                )

            if pkey.is_absolute():
                dset = fid[pkey.parent].createVariable(
                    pkey.name,
                    value["_dtype"],
                    dimensions=(pkey.name,),
                    fill_value=fillvalue,
                    contiguous=value["_size"] != 0,
                )
            else:
                dset = fid.createVariable(
                    key,
                    value["_dtype"],
                    dimensions=(key,),
                    fill_value=fillvalue,
                    contiguous=value["_size"] != 0,
                )
            if value["_size"] > 0:
                if "_values" in value:
                    dset[:] = np.array(value["_values"])
                elif "_range" in value:
                    dset[:] = np.arange(*value["_range"], dtype=value["_dtype"])

            dset.setncatts(
                {
                    k: adjust_attr(value["_dtype"], k, v)
                    for k, v in value.items()
                    if not k.startswith("_")
                }
            )

    def __compounds(self: NcYaml, fid: Dataset) -> None:
        """Add compound datatypes to a netCDF4 product.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """
        for key, val in self.nc_def["compounds"].items():
            comp_t = np.dtype([(k, v[0]) for k, v in val.items()])
            _ = fid.createCompoundType(comp_t, key)

    def __variables(self: NcYaml, fid: Dataset) -> None:
        """Add datasets to a netCDF4 product.

        Parameters
        ----------
        fid :  netCDF4.Dataset
           netCDF4 Dataset (mode 'r+')

        """

        def add_compound_attr() -> None:
            compound = self._nc_def["compounds"][val["_dtype"]]
            res = [v[2] for k, v in compound.items() if len(v) == 3]
            if res:
                dset.units = [v[1] for k, v in compound.items()]
                dset.names = res
            else:
                dset.names = [v[1] for k, v in compound.items()]

        for key, val in self.nc_def["variables"].items():
            pkey = PurePosixPath(key)
            var_grp = fid[pkey.parent] if pkey.is_absolute() else fid
            var_name = pkey.name if pkey.is_absolute() else key

            if val["_dtype"] in fid.cmptypes:
                is_compound = True
                ds_dtype = fid.cmptypes[val["_dtype"]]
            else:
                is_compound = False
                ds_dtype = val["_dtype"]

            fillvalue = None
            if "_FillValue" in val:
                fillvalue = (
                    np.nan if val["_FillValue"] == "NaN" else int(val["_FillValue"])
                )

            # check for scalar dataset
            if val["_dims"][0] == "scalar":
                dset = var_grp.createVariable(
                    var_name,
                    ds_dtype,
                    fill_value=fillvalue,
                    contiguous=True,
                )
                dset.setncatts(
                    {
                        k: adjust_attr(ds_dtype, k, v)
                        for k, v in val.items()
                        if not k.startswith("_")
                    }
                )
                if is_compound:
                    add_compound_attr()
                continue

            compression = None
            complevel = 0
            # currently only gzip compression is supported
            if "_compression" in val:
                compression = "zlib"
                complevel = val["_compression"]

            var_dims = []
            n_udim = 0
            ds_shape = ()
            ds_maxshape = ()
            for coord in val["_dims"]:
                pcoord = PurePosixPath(coord)
                var_dims.append(pcoord.name if pcoord.is_absolute() else coord)
                if pcoord.is_absolute():
                    dim_sz = fid[pcoord.parent].dimensions[pcoord.name].size
                else:
                    dim_sz = fid.dimensions[coord].size
                n_udim += int(dim_sz == 0)
                ds_shape += (dim_sz,)
                ds_maxshape += (dim_sz if dim_sz > 0 else None,)

            # currently, we can not handle more than one unlimited dimension
            if n_udim > 1:
                raise ValueError("more than one unlimited dimension")

            if None in ds_maxshape and val.get("_chunks") == "contiguous":
                raise KeyError(
                    "you can not create a contiguous dataset with unlimited dimensions."
                )

            # create the variable
            if val.get("_chunks") == "contiguous":
                dset = var_grp.createVariable(
                    var_name,
                    ds_dtype,
                    dimensions=var_dims,
                    fill_value=fillvalue,
                    contiguous=True,
                )
            else:
                ds_chunk = val.get("_chunks")
                if ds_chunk is not None and not isinstance(ds_chunk, bool):
                    ds_chunk = tuple(ds_chunk)
                if val.get("_vlen"):
                    if is_compound:
                        raise ValueError("can not have vlen with compounds")
                    ds_dtype = fid.createVLType(ds_dtype, "phony_vlen")
                    fillvalue = None

                dset = var_grp.createVariable(
                    var_name,
                    str if val["_dtype"] == "str" else ds_dtype,
                    dimensions=var_dims,
                    fill_value=fillvalue,
                    compression=compression,
                    complevel=complevel,
                    chunksizes=ds_chunk,
                    contiguous=False,
                )
            dset.setncatts(
                {
                    k: adjust_attr(val["_dtype"], k, v)
                    for k, v in val.items()
                    if not k.startswith("_")
                }
            )
            if is_compound:
                add_compound_attr()

    @property
    def nc_def(self: NcYaml) -> dict:
        """Return definition of the netCDF4 product."""
        return self._nc_def

    def diskless(self: NcYaml, persist: bool = False) -> Dataset:
        """Create a netCDF4 file in memory."""
        fid = Dataset("diskless_test.nc", "w", diskless=True, persist=persist)
        self.__groups(fid)
        self.__dimensions(fid)
        self.__compounds(fid)
        self.__variables(fid)
        self.__attrs(fid)
        return fid

    def create(self: NcYaml, l1a_name: Path | str) -> None:
        """Create a netCDF4 file (overwrite if exist).

        Parameters
        ----------
        l1a_name :  Path | str
           Full name of the netCDF4 file to be generated

        """
        try:
            with Dataset(l1a_name, "w") as fid:
                self.__groups(fid)
                self.__dimensions(fid)
                self.__compounds(fid)
                self.__variables(fid)
                self.__attrs(fid)
        except PermissionError as exc:
            raise RuntimeError(f"failed to create {l1a_name}") from exc



#
# This file is part of Python package: `h5yaml`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025 - R.M. van Hees (SRON)
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test module for h5yaml function `conf_from_yaml`."""

from __future__ import annotations

from importlib.resources import files

import pytest

from h5yaml.conf_from_yaml import conf_from_yaml


def test_from_yaml() -> None:
    """..."""
    with pytest.raises(FileNotFoundError, match="not found") as excinfo:
        _ = conf_from_yaml(files("h5yaml.Data") / "not_existing.yaml")
    assert f"{files('h5yaml.Data') / 'not_existing.yaml'} not found" in str(excinfo)

    with pytest.raises(RuntimeError, match=r"Failed to parse .*") as excinfo:
        _ = conf_from_yaml("README.md")
    assert "Failed to parse" in str(excinfo)

    res = conf_from_yaml(files("h5yaml.Data") / "nc_testing.yaml")
    assert isinstance(res, dict)
